@@ -1,11 +1,12 @@
 /*
  * wcsrtombs.c
  *
- * MinGW.org replacement for the wcsrtombs() function; delegates to the
- * Microsoft implementation, if available in the C runtime DLL, otherwise
- * handles the call locally.
+ * MinGW.org implementation of the wcsrtombs() function; supports use of
+ * this function on any legacy Windows version, for which Microsoft do not
+ * provide it, and replaces the Microsoft implementation, when they do.
  *
- * $Id: wcsrtombs.c,v 4553677b3e4e 2020/04/02 19:37:13 keith $
+ *
+ * $Id: wcsrtombs.c,v 3346bd5e5239 2020/07/04 21:06:58 keith $
  *
  * Written by Keith Marshall <keith@users.osdn.me>
  * Copyright (C) 2019, 2020, MinGW.org Project
@@ -33,19 +34,12 @@
  */
 #include "wcharmap.h"
 
-/* For runtime delegation, we need a mechanism for detection of an
- * implementation, within the default C runtime DLL; we may use the
- * MinGW dlfcn emulation, to facilitate this.
- */
-#include <dlfcn.h>
-
-static size_t __mingw_wcsrtombs_fallback
+static __mb_inline__ size_t __mingw_wcsrtombs_internal
 ( char *restrict mbs, const wchar_t **restrict wcs, size_t len,
   mbstate_t *restrict ps
 )
-{ /* Fallback function, providing an implementation of the wcsrtombs()
-   * function, when none is available within the Microsoft C runtime, or
-   * the user has elected to override any such Microsoft implementation.
+{ /* Internal implementation of the wcsrtombs() function; this will be
+   * expanded inline, within the body of the public implementation.
    *
    * Initially, save the current errno state, so that we may restore
    * it on return, clear it to zero for internal checking, and compute
@@ -144,63 +138,30 @@ static size_t __mingw_wcsrtombs_fallback
   return errout( errno_reset, count );
 }
 
-size_t __mingw_wcsrtombs
-( char *mbs, const wchar_t **wcs, size_t len, mbstate_t *ps )
+size_t wcsrtombs( char *mbs, const wchar_t **wcs, size_t len, mbstate_t *ps )
 {
-  /* Wrapper for the wcsrtombs() function; this will unconditionally
-   * delegate the call to the MinGW fallback implementation, (defined
-   * above), after first ensuring that the specified wcs reference is
-   * valid, and that the effective codeset has been initialized.
+  /* Implementation of ISO-C99 wcsrtombs() function, in libmingwex.a;
+   * before proceeding, we must ensure that the wcs argument specifies
+   * an indirect reference to a non-NULL wchar_t array.
    */
   if( (wcs == NULL) || (*wcs == NULL) ) return errout( EINVAL, (size_t)(-1) );
 
+  /* With a valid wcs reference, store the effective codeset, and
+   * hand the conversion off to the inline expansion of the preceding
+   * implementation.
+   */
   (void)(__mingw_wctomb_codeset_init() );
-  return __mingw_wcsrtombs_fallback( mbs, wcs, len, ps );
+  return __mingw_wcsrtombs_internal( mbs, wcs, len, ps );
 }
 
-size_t __msvcrt_wcsrtombs
-( char *mbs, const wchar_t **wcs, size_t len, mbstate_t *ps )
-{
-  /* Wrapper for the wcsrtombs() function; it will initially attempt
-   * to delegate the call to a Microsoft-provided implementation, but
-   * if no such implementation can be found, fall back to the MinGW
-   * substitute (defined above).
-   */
-  typedef size_t (*redirect_t)(char *, const wchar_t **, size_t, mbstate_t *);
-  static redirect_t redirector_hook = NULL;
+/* FIXME: these aliases are provided for link-compatibitity with
+ * libraries compiled against mingwrt-5.3.x; they may be removed
+ * from future versions of mingwrt.
+ */
+size_t __msvcrt_wcsrtombs( char *, const wchar_t **, size_t, mbstate_t * )
+__attribute__((__weak__,__alias__("wcsrtombs")));
 
-  /* Neither wcs, not the pointer to which it refers, may be NULL.
-   * ISO C doesn't specify any particular outcome for this condition,
-   * (so a segmentation fault would conform); it makes more sense to
-   * catch the abnormality, and bail out.
-   */
-  if( (wcs == NULL) || (*wcs == NULL) ) return errout( EINVAL, (size_t)(-1) );
-
-  /* MSVCRT.DLL's setlocale() cannot reliably handle code pages with
-   * more than two bytes per code point, (e.g. UTF-7 and UTF-8); thus,
-   * Microsoft's wcsrtombs() is likely to be similarly unreliable, so
-   * always use the MinGW fallback with such code pages.
-   */
-  if( __mb_cur_max_for_codeset( __mingw_wctomb_codeset_init() ) > 2 )
-    return __mingw_wcsrtombs_fallback( mbs, wcs, len, ps );
-
-  /* On first time call, we don't know which implementation is to be
-   * selected; look for a Microsoft implementation, which, if available,
-   * may be registered for immediate use on this, and any subsequent,
-   * calls to this function wrapper...
-   */
-  if(  (redirector_hook == NULL)
-  &&  ((redirector_hook = dlsym( RTLD_DEFAULT, "wcsrtombs" )) == NULL)  )
-  {
-    /* ...but when no Microsoft implementation can be found, register
-     * the MinGW fallback in its stead.
-     */
-    redirector_hook = __mingw_wcsrtombs_fallback;
-  }
-  /* Finally, delegate the call to whichever implementation has been
-   * registered on first-time call.
-   */
-  return redirector_hook( mbs, wcs, len, ps );
-}
+size_t __mingw_wcsrtombs( char *, const wchar_t **, size_t, mbstate_t * )
+__attribute__((__weak__,__alias__("wcsrtombs")));
 
 /* $RCSfile: wcsrtombs.c,v $: end of file */
